@@ -47,7 +47,9 @@ use AI::MXNet::Contrib;
 use AI::MXNet::CachedOp;
 use AI::MXNet::AutoGrad;
 use AI::MXNet::Gluon;
-our $VERSION = '1.1';
+use AI::MXNet::NDArray::Sparse;
+use AI::MXNet::Symbol::Sparse;
+our $VERSION = '1.2';
 
 sub import
 {
@@ -78,6 +80,7 @@ sub import
             sub io { 'AI::MXNet::IO' }
             sub metric { 'AI::MXNet::Metric' }
             sub mod { 'AI::MXNet::Module' }
+            sub module { 'AI::MXNet::Module' }
             sub mon { 'AI::MXNet::Monitor' }
             sub viz { 'AI::MXNet::Visualization' }
             sub rnn { 'AI::MXNet::RNN' }
@@ -196,8 +199,6 @@ AI::MXNet - Perl interface to MXNet machine learning library
 
 =head1 COPYRIGHT & LICENSE
 
-    Copyright (C) 2017 by Sergey Kolychev <sergeykolychev.github@gmail.com>
-
     This library is licensed under Apache 2.0 license https://www.apache.org/licenses/LICENSE-2.0
 
 =cut
