@@ -19,21 +19,26 @@ package AI::MXNet::Base;
 use strict;
 use warnings;
 use PDL;
-use PDL::Types qw();
-use AI::MXNetCAPI 1.1;
-use AI::NNVMCAPI 1.1;
+use PDL::Types ();
+use PDL::CCS::Nd;
+use AI::MXNetCAPI 1.2;
+use AI::NNVMCAPI 1.2;
 use AI::MXNet::Types;
 use Time::HiRes;
+use Scalar::Util qw(blessed);
 use Carp;
 use Exporter;
 use base qw(Exporter);
 use List::Util qw(shuffle);
 
 @AI::MXNet::Base::EXPORT = qw(product enumerate assert zip check_call build_param_doc
-                              pdl cat dog svd bisect_left pdl_shuffle as_array
+                              pdl cat dog svd bisect_left pdl_shuffle as_array ascsr rand_sparse
                               DTYPE_STR_TO_MX DTYPE_MX_TO_STR DTYPE_MX_TO_PDL
-                              DTYPE_PDL_TO_MX DTYPE_MX_TO_PERL GRAD_REQ_MAP);
-@AI::MXNet::Base::EXPORT_OK = qw(pzeros pceil);
+                              DTYPE_PDL_TO_MX DTYPE_MX_TO_PERL GRAD_REQ_MAP
+                              STORAGE_TYPE_UNDEFINED STORAGE_TYPE_DEFAULT
+                              STORAGE_TYPE_ROW_SPARSE STORAGE_TYPE_CSR
+                              STORAGE_TYPE_STR_TO_ID STORAGE_TYPE_ID_TO_STR STORAGE_AUX_TYPES);
+@AI::MXNet::Base::EXPORT_OK = qw(pzeros pceil pones digitize hash array_index range);
 
 use constant DTYPE_STR_TO_MX => {
     float32 => 0,
@@ -97,6 +102,29 @@ use constant GRAD_REQ_MAP => {
     write => 1,
     add   => 3
 };
+use constant {
+    STORAGE_TYPE_UNDEFINED  => -1,
+    STORAGE_TYPE_DEFAULT    =>  0,
+    STORAGE_TYPE_ROW_SPARSE =>  1,
+    STORAGE_TYPE_CSR        =>  2
+};
+use constant STORAGE_TYPE_STR_TO_ID => {
+    undefined  => STORAGE_TYPE_UNDEFINED,
+    default    => STORAGE_TYPE_DEFAULT,
+    row_sparse => STORAGE_TYPE_ROW_SPARSE,
+    csr        => STORAGE_TYPE_CSR
+};
+use constant STORAGE_TYPE_ID_TO_STR => {
+    STORAGE_TYPE_UNDEFINED()  => 'undefined',
+    STORAGE_TYPE_DEFAULT()    => 'default',
+    STORAGE_TYPE_ROW_SPARSE() => 'row_sparse',
+    STORAGE_TYPE_CSR()        => 'csr'
+};
+use constant STORAGE_AUX_TYPES => {
+    row_sparse => ['int64'],
+    csr => ['int64', 'int64']
+};
+
 
 =head1 NAME
 
@@ -120,12 +148,17 @@ use constant GRAD_REQ_MAP => {
 
 sub zip
 {
-    my ($sub, @arrays) = @_;
-    my $len = @{ $arrays[0] };
-    for (my $i = 0; $i < $len; $i++)
+    if('CODE' eq ref $_[0])
     {
-        $sub->(map { $_->[$i] } @arrays);
+        # continue supporting the callback style
+        my $code = shift;
+        $code->(@$_) for AI::MXNetCAPI::py_zip(map { \@$_ } @_);
+        return;
     }
+    # the map() here may seem like a no-op, but triggers overloading or
+    # whatever else is needed to make array-ish things actually arrays
+    # before entering the low level list builder.
+    return AI::MXNetCAPI::py_zip(map { \@$_ } @_);
 }
 
 =head2 enumerate
@@ -193,18 +226,11 @@ sub bisect_left
     as shuffled last dimension's indexes
 =cut
 
-
 sub pdl_shuffle
 {
     my ($pdl, $preshuffle) = @_;
-    my $c = $pdl->copy;
     my @shuffle = $preshuffle ? @{ $preshuffle } : shuffle(0..$pdl->dim(-1)-1);
-    my $rem = $pdl->ndims-1;
-    for my $i (0..$pdl->dim(-1)-1)
-    {
-        $c->slice(('X')x$rem, $i) .= $pdl->slice(('X')x$rem, $shuffle[$i])
-    }
-    $c;
+    return $pdl->dice_axis(-1, pdl(\@shuffle));
 }
 
 =head2 assert
@@ -270,16 +296,14 @@ sub build_param_doc
     $remove_dup //= 1;
     my %param_keys;
     my @param_str;
-    zip(sub {
-            my ($key, $type_info, $desc) = @_;
-            return if exists $param_keys{$key} and $remove_dup;
+    for(zip($arg_names, $arg_types, $arg_descs)) {
+            my ($key, $type_info, $desc) = @$_;
+            next if exists $param_keys{$key} and $remove_dup;
             $param_keys{$key} = 1;
             my $ret = sprintf("%s : %s", $key, $type_info);
             $ret .= "\n    ".$desc if length($desc);
             push @param_str,  $ret;
-        },
-        $arg_names, $arg_types, $arg_descs
-    );
+    }
     return sprintf("Parameters\n----------\n%s\n", join("\n", @param_str));
 }
 
@@ -345,10 +369,76 @@ END {
 }
 
 *pzeros = \&zeros;
+*pones = \&ones;
 *pceil  = \&ceil;
 ## making sure that we can stringify arbitrarily large piddles
 $PDL::toolongtoprint = 1000_000_000;
 ## convenience subs
+
+sub ascsr
+{
+    my ($data, $indptr, $indices, $shape) = @_;
+    my @which;
+    my $i = 0;
+    my $j = 0;
+    while($i < $indices->nelem)
+    {
+        for($i = $indptr->at($j); $i < $indptr->at($j+1); $i++)
+        {
+            push @which, [$j, $indices->at($i)];
+        }
+        $j++;
+    }
+    return PDL::CCS::Nd->newFromWhich(
+            pdl(\@which), $data, pdims => blessed $shape ? $shape : pdl($shape)
+    )->xchg(0, 1);
+}
+
+package AI::MXNet::COO::Nd;
+use Mouse;
+has ['data', 'row', 'col'] => (is => 'rw');
+no Mouse;
+
+package AI::MXNet::Base;
+
+sub tocoo
+{
+    my $csr = shift;
+    return AI::MXNet::COO::Nd->new(
+        data => $csr->data,
+        row  => $csr->_whichND->slice(0)->flat,
+        col  => $csr->_whichND->slice(1)->flat
+    );
+}
+
+sub rand_sparse
+{
+    my ($num_rows, $num_cols, $density, $dtype, $format) = @_;
+    $dtype  //= 'float32';
+    $format //= 'csr';
+    my $pdl_type = PDL::Type->new(DTYPE_MX_TO_PDL->{ $dtype });
+    my $dense = random($pdl_type, $num_cols, $num_rows);
+    my $missing = 0;
+    $dense->where(random($num_cols, $num_rows)<=1-$density) .= $missing;
+    if($format eq 'csr')
+    {
+        return $dense->tocsr;
+    }
+    return $dense;
+}
+
+{
+    no warnings 'once';
+    *PDL::CCS::Nd::data    = sub { shift->_nzvals };
+    *PDL::CCS::Nd::indptr  = sub { my $self = shift; ($self->hasptr ? $self->getptr : $self->ptr)[0] };
+    *PDL::CCS::Nd::indices = sub { shift->_whichND->slice(1)->flat };
+    *PDL::CCS::Nd::tocoo   = sub { tocoo(shift) };
+    *PDL::CCS::Nd::shape   = sub { shift->pdims };
+    *PDL::CCS::Nd::dtype   = sub { DTYPE_MX_TO_STR->{ DTYPE_PDL_TO_MX->{ shift->type->numval } } };
+    *PDL::tocsr            = sub { shift->xchg(0, 1)->toccs->xchg(0, 1) };
+    *PDL::rand_sparse      = sub { shift; rand_sparse(@_) };
+}
+
 {
     my $orig_at = PDL->can('at');
     no warnings 'redefine';
@@ -362,4 +452,20 @@ $PDL::toolongtoprint = 1000_000_000;
     *PDL::dtype  = sub { DTYPE_MX_TO_STR->{ DTYPE_PDL_TO_MX->{ shift->type->numval } } };
 }
 
+sub digitize
+{
+    my ($d, $bins) = @_;
+    for(my $i = 0; $i < @$bins; $i++)
+    {
+        return $i if $d < $bins->[$i];
+    }
+    return scalar(@$bins);
+}
+
+use B;
+sub hash { hex(B::hash(shift)) }
+use List::Util ();
+sub array_index { my ($s, $array) = @_; return List::Util::first { $array->[$_] eq $s } 0..@$array-1 }
+sub range { my ($begin, $end, $step) = @_; $step //= 1; grep { not (($_-$begin) % $step) } $begin..$end-1 }
+
 1;
